@@ -31,6 +31,18 @@ class InvertedLogicException(Exception):
         return self.message
 
 
+class InvalidErrorRateException(Exception):
+    def __init__(self, *, falsePositiveRate):
+        self.message = (
+            "All false positive rates must be between 0 and 1, but this one is"
+            f"{falsePositiveRate}. Consider calculating the error rate using the "
+            "set_crlite_error_rates function."
+        )
+
+    def __str__(self):
+        return self.message
+
+
 # A simple-as-possible bloom filter implementation making use of version 3 of the 32-bit murmur
 # hash function (for compat with multi-level-bloom-filter-js).
 # mgoodwin 2018
@@ -135,12 +147,19 @@ class Bloomer:
 
     @classmethod
     def calc_n_hashes(cls, falsePositiveRate):
-        return math.ceil(math.log(1.0 / falsePositiveRate, 2))
+        if falsePositiveRate < 0 or falsePositiveRate > 1:
+            raise InvalidErrorRateException(falsePositiveRate=falsePositiveRate)
+        nHashes = math.ceil(math.log2(1.0 / falsePositiveRate))
+        assert nHashes > 0, "Always must have a positive number of hashes"
+        return nHashes
 
     @classmethod
     def calc_size(cls, nHashFuncs, elements, falsePositiveRate):
+        if falsePositiveRate < 0 or falsePositiveRate > 1:
+            raise InvalidErrorRateException(falsePositiveRate=falsePositiveRate)
         # From CRLite paper, https://cbw.sh/static/pdf/larisch-oakland17.pdf
-        min_bits = math.ceil(1.44 * elements * math.log(1 / falsePositiveRate, 2))
+        min_bits = math.ceil(1.44 * elements * math.log2(1 / falsePositiveRate))
+        assert min_bits > 0, "Always must have a positive number of bits"
         # Ensure the result is divisible by 8 for full bytes
         return 8 * math.ceil(min_bits / 8)
 
@@ -176,10 +195,11 @@ class Bloomer:
 class FilterCascade:
     def __init__(
         self,
-        filters,
+        filters=None,
+        *,
         error_rates=[0.02, 0.5],
         growth_factor=1.1,
-        min_filter_length=10_000,
+        min_filter_length=1_000,
         version=2,
         defaultHashAlg=fileformats.HashAlgorithm.MURMUR3,
         salt=None,
@@ -191,8 +211,7 @@ class FilterCascade:
             invertedLogic: If not supplied (or left as None), it will be auto-
                 detected.
         """
-        self.filters = filters
-        self.error_rates = error_rates
+        self.filters = filters or []
         self.growth_factor = growth_factor
         self.min_filter_length = min_filter_length
         self.version = version
@@ -200,6 +219,8 @@ class FilterCascade:
         self.salt = salt
         self.invertedLogic = invertedLogic
 
+        self.set_error_rates(error_rates)
+
         if self.salt and version < 2:
             raise ValueError("salt requires format version 2 or greater")
         if self.salt and not isinstance(self.salt, bytes):
@@ -211,13 +232,21 @@ class FilterCascade:
         if self.salt and self.defaultHashAlg == fileformats.HashAlgorithm.MURMUR3:
             raise ValueError("salts not permitted for MurmurHash3")
 
+    def set_error_rates(self, new_rates):
+        if len(new_rates) < 1:
+            raise ValueError("You must supply at least one error rate")
+        for falsePositiveRate in new_rates:
+            if falsePositiveRate <= 0 or falsePositiveRate >= 1:
+                raise InvalidErrorRateException(falsePositiveRate=falsePositiveRate)
+        self.error_rates = new_rates
+        return self.error_rates
+
     def set_crlite_error_rates(self, *, include_len, exclude_len):
         if include_len > exclude_len:
             raise InvertedLogicException(
                 depth=None, exclude_count=exclude_len, include_len=include_len
             )
-        self.error_rates = [include_len / (math.sqrt(2) * exclude_len), 0.5]
-        return self.error_rates
+        return self.set_error_rates([include_len / (math.sqrt(2) * exclude_len), 0.5])
 
     def initialize(self, *, include, exclude):
         """
@@ -258,8 +287,6 @@ class FilterCascade:
 
             if depth > len(self.filters):
                 self.filters.append(
-                    # For growth-stability reasons, we force all layers to be at least
-                    # min_filter_length large. This is important for the deep layers near the end.
                     Bloomer.filter_with_characteristics(
                         elements=max(
                             int(include_len * self.growth_factor),
@@ -279,18 +306,28 @@ class FilterCascade:
                 if self.filters[depth - 1].size < required_size:
                     # Resize filter
                     self.filters[depth - 1] = Bloomer.filter_with_characteristics(
-                        int(include_len * self.growth_factor), er, depth
+                        elements=int(include_len * self.growth_factor),
+                        falsePositiveRate=er,
+                        level=depth,
+                        hashAlg=self.defaultHashAlg,
                     )
-                    log.info("Resized filter at {}-depth layer".format(depth))
-            filter = self.filters[depth - 1]
+                    log.info(
+                        f"Resized filter at {depth}-depth layer to {self.filters[depth - 1].size}"
+                    )
+
+            current_filter = self.filters[depth - 1]
             log.debug(
                 "Initializing the {}-depth layer. err={} include_len={} size={} hashes={}".format(
-                    depth, er, include_len, filter.size, filter.nHashFuncs
+                    depth,
+                    er,
+                    include_len,
+                    current_filter.size,
+                    current_filter.nHashFuncs,
                 )
             )
             # loop over the elements that *should* be there. Add them to the filter.
             for elem in include:
-                filter.add(elem)
+                current_filter.add(elem)
 
             # loop over the elements that should *not* be there. Create a new layer
             # that *includes* the false positives and *excludes* the true positives
@@ -299,10 +336,11 @@ class FilterCascade:
             exclude_count = 0
             for elem in exclude:
                 exclude_count += 1
-                if elem in filter:
+                if elem in current_filter:
                     false_positives.add(elem)
+            log.debug(f"Found {len(false_positives)} false positives at depth {depth}.")
 
-            if exclude_count < include_len:
+            if depth == 1 and exclude_count < include_len:
                 raise InvertedLogicException(
                     depth=depth, exclude_count=exclude_count, include_len=include_len
                 )
@@ -313,19 +351,19 @@ class FilterCascade:
                     (endtime - starttime).seconds * 1000
                     + (endtime - starttime).microseconds / 1000,
                     depth,
-                    len(filter.bitarray),
+                    len(current_filter.bitarray),
                 )
             )
             # Sanity check layer growth.  Bit count should be going down
             # as false positive rate decreases.
             if depth > 2:
-                if len(filter.bitarray) > len(self.filters[depth - 3].bitarray):
+                if len(current_filter.bitarray) > len(self.filters[depth - 3].bitarray):
                     sequentialGrowthLayers += 1
                     log.warning(
                         "Increase in false positive rate detected. Depth {} has {}"
                         " bits and depth {} has {} bits. {}/{} allowed warnings.".format(
                             depth,
-                            len(filter.bitarray),
+                            len(current_filter.bitarray),
                             depth - 3 + 1,
                             len(self.filters[depth - 3].bitarray),
                             sequentialGrowthLayers,
@@ -374,9 +412,9 @@ class FilterCascade:
 
     def verify(self, *, include, exclude):
         for entry in include:
-            assert entry in self, "oops! false negative!"
+            assert entry in self, f"Verification Failure: false negative: {entry}"
         for entry in exclude:
-            assert entry not in self, "oops! false positive!"
+            assert entry not in self, f"Verification Failure: false positive: {entry}"
 
     def bitCount(self):
         total = 0
@@ -429,7 +467,7 @@ class FilterCascade:
 
         filters = []
         while len(buf) > 0:
-            (buf, f) = Bloomer.from_buf(buf)
+            (buf, f) = Bloomer.from_buf(buf, salt=salt)
             filters.append(f)
         assert len(buf) == 0, "buffer should be consumed"
 
@@ -442,7 +480,7 @@ class FilterCascade:
                 raise ValueError("All filter layers should use the same hash algorithm")
 
         return FilterCascade(
-            filters,
+            filters=filters,
             version=version,
             defaultHashAlg=hashAlg,
             salt=salt,
@@ -450,6 +488,13 @@ class FilterCascade:
         )
 
     @classmethod
+    @deprecated(
+        version="0.4.0",
+        reason=(
+            "Use the standard __init__ function generally set error rates with "
+            + "set_crlite_error_rates or set_error_rates"
+        ),
+    )
     def cascade_with_characteristics(
         cls,
         *,
@@ -460,11 +505,12 @@ class FilterCascade:
         layer=0,
     ):
         return FilterCascade(
-            [
+            filters=[
                 Bloomer.filter_with_characteristics(
                     elements=capacity,
                     falsePositiveRate=error_rates[0],
                     hashAlg=defaultHashAlg,
+                    salt=salt,
                 )
             ],
             error_rates=error_rates,
